/* 
 * Copyright (C) 2008 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "boxeditor.h"
#include "util.h"
#include "bittwiddler.h"
#include "tessboxes.h"

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <signal.h> /* defines SIGWINCH equals 28 */

#include <slang.h>
#include <readline/readline.h>
#include <readline/history.h> 

display_state_t disp; /* global for signal handlers */

extern options_t u_options;

static void sigwinch_handler (int sig) {

  SLsig_block_signals();

  SLtt_get_screen_size();
  SLsmg_reinit_smg();

  if( !disp.delay_sigwinch ) {
    redraw_boxeditor(disp.ed);
  }

  SLsig_unblock_signals();

  SLsignal(SIGWINCH, sigwinch_handler);

}

void init_boxeditor(int *argc, char ***argv) {
}

bool_t create_boxeditor(boxeditor_t *e, imagebuffer_t *ib, boxlist_t *bl) {
  int i;

  if( e ) {

    e->bl = bl;
    e->ib = ib;
    e->ds = &disp;
    e->dirty_list = TRUE;
    e->dirty_graph = FALSE;

    if( !create_xtwin(&e->xtw, ib->width, ib->height) ) {
      fprintf (stderr, "Unable to initialize terminal.");
      return FALSE;
    }

    SLang_init_tty(-1, 0, 0);
    SLtt_get_terminfo();
    if(-1 == SLsmg_init_smg() ) {
      fprintf (stderr, "Unable to initialize terminal.");
      return FALSE;
    }

    SLkp_init(); /* for cursor keys */

    SLsignal(SIGWINCH, sigwinch_handler);

    /* this lets us print utf8 properly */
    SLutf8_enable(1);
    SLtt_utf8_enable(1);
    SLsmg_utf8_enable(1);

    /* initialize the display state */
    disp.first_visible = 0;
    disp.highlighted = 0;

    SLtt_set_color(0, NULL, "lightgray", "default");
    SLtt_set_color(1, NULL, "black", "lightgray");
    SLtt_set_color(2, NULL, "lightgray", "blue");
    SLtt_set_color(3, NULL, "white", "red");
    SLtt_set_color(4, NULL, "default", "lightgray");

    SLtt_set_cursor_visibility(0);

    for(i = 0; i < 10; i++) {
      disp.fkey_cmd[i] = NULL;
    }

    return TRUE;
  }
  return FALSE;
}

bool_t free_boxeditor(boxeditor_t *e) {
  if( e ) {
    free_xtwin(&e->xtw);
    /* we're done */
    SLtt_set_cursor_visibility(1);
    SLsmg_reset_smg();
    SLang_reset_tty();

    e->bl = NULL;
    e->ib = NULL;
    e->ds = NULL;
    disp.ed = NULL;
  }
  return FALSE;
}

void render_text_boxeditor(boxeditor_t *e) {
  int r;
  box_t *f;
  int w = 0;
  int c = 0;
  bool_t space;

  SLsmg_set_color(4);
  for(r = 0; 
      (r + disp.first_visible < e->bl->count) && (c < disp.num_cols);
      r++) {
    f = &(e->bl->items[r + disp.first_visible]);

    SLsmg_normal_video();
    /* this is a very simple method to identify word separators,
       and if you're reading this it's probably no good :) */
    space = ( f->left - w > f->right - f->left ) ? TRUE : FALSE;
    
    if( space ) {
      SLsmg_write_char(L' ');
    }

    if(r == disp.highlighted) {
      SLsmg_reverse_video();
    }

    SLsmg_printf("%s", f->symbol);

    w = f->right;
    c += (space ? 1 : 0) + SLutf8_strlen(f->symbol,FALSE);
  }
}


void redraw_boxeditor(boxeditor_t *e) {
  int r;
  box_t *f;

  SLsig_block_signals();

  if( e ) {

    disp.num_rows = SLtt_Screen_Rows;
    disp.num_cols = SLtt_Screen_Cols;

    SLsmg_normal_video();

    if( e->bl->count > 0 ) {
      for(r = 0; r < (disp.num_rows - 3); r++) {
	if( r + disp.first_visible < e->bl->count ) {

	  f = &(e->bl->items[r + disp.first_visible]);
	  
	  /* this can't be done with a single printf format
	     string, because utf character sizes will muck
	     it up */
	  SLsmg_gotorc(r + 1, 0);
	  SLsmg_printf(" %-8s", f->symbol);
	  SLsmg_erase_eol(); 
	  SLsmg_refresh();
	  SLsmg_gotorc(r + 1, 16);
	  SLsmg_printf("%6u\t%6u\t%6u\t%6u",
		       f->left, f->top, f->right, f->bottom);
	  SLsmg_erase_eol();
	} else {
	  SLsmg_gotorc(r + 1, 0);
	  SLsmg_erase_eol();
	}
      }
    } else {
      SLsmg_cls();
      r = disp.num_rows - 3;
    }

    SLsmg_gotorc(r + 1, 0);
    render_text_boxeditor(e);
    SLsmg_erase_eol();

    SLsmg_reverse_video();

    f = &(e->bl->items[disp.first_visible + disp.highlighted]);

    SLsmg_gotorc(disp.highlighted + 1, 0);

    if( e->bl->count > 0 ) {
      SLsmg_printf(" %-8s", f->symbol);
      SLsmg_erase_eol();
      SLsmg_refresh();
      SLsmg_gotorc(disp.highlighted + 1, 16);
      SLsmg_printf("%6u\t%6u\t%6u\t%6u",
		   f->left, f->top, f->right, f->bottom);
    }
    SLsmg_erase_eol();

    SLsmg_set_color(2);

    SLsmg_gotorc(0,0);
    SLsmg_printf("tessboxes " PACKAGE_VERSION);
    SLsmg_erase_eol();

    SLsmg_gotorc(r + 2, 0);
    SLsmg_printf("Ctrl+x: quit Alt+arrows: move Ctrl+arrows: grow/shrink Other keys: see man");
    SLsmg_erase_eol();
    SLsmg_refresh();

  }

  SLsig_unblock_signals();

}

/* Every terminal generates a sequence of bytes in response
   to one or more simultaneous keypresses. The values below
   have been obtained by inspection from a handful of terminals.
   They work on my machine, but it's not a fully portable way to
   do it. If your terminal sends different codes, just add them
   in the list. See the function next_key_code()
*/
#define KCLEN 8
typedef int keycode_t[KCLEN];
typedef struct {
  int sym;
  keycode_t code;
} keysym_t;
#define kESC      0
#define kUP       1
#define kDOWN     2
#define kLEFT     3
#define kRIGHT    4
#define kALTUP    5
#define kALTDOWN  6
#define kALTLEFT  7
#define kALTRIGHT 8
#define kCTLUP    9
#define kCTLDOWN  10
#define kCTLLEFT  11
#define kCTLRIGHT 12
#define kCTLX     13
#define kINS      14
#define kDEL      15
#define kPGUP     16
#define kPGDOWN   17
#define kHOME     18
#define kEND      19
#define kNORMAL   20
#define kNOTHING  21
#define kALTC     22
#define kALTA     23
#define kALTS     24
#define kMULTIB   25
#define kPREV     26

static keysym_t keysyms[] = {
  { kESC, { 27, 0, 0, 0, 0, 0, 0, 0 }},
  { kUP, { 27, 91, 65, 0, 0, 0, 0, 0}},
  { kDOWN, { 27, 91, 66, 0, 0, 0, 0, 0 }},
  { kLEFT, { 27, 91, 68, 0, 0, 0, 0, 0 }}, 
  { kRIGHT, { 27, 91, 67, 0, 0, 0, 0, 0 }}, 
  { kALTUP, { 27, 27, 91, 65, 0, 0, 0, 0 }}, 
  { kALTUP, { 27, 91, 49, 59, 51, 65, 0, 0 }}, 
  { kALTDOWN, { 27, 27, 91, 66, 0, 0, 0, 0 }}, 
  { kALTDOWN, { 27, 91, 49, 59, 51, 66, 0, 0 }}, 
  { kALTLEFT, { 27, 27, 91, 68, 0, 0, 0, 0 }}, 
  { kALTLEFT, { 27, 91, 49, 59, 51, 68, 0, 0 }}, 
  { kALTRIGHT, { 27, 27, 91, 67, 0, 0, 0, 0 }}, 
  { kALTRIGHT, { 27, 91, 49, 59, 51, 67, 0, 0 }}, 
  { kCTLUP, { 27, 79, 97, 0, 0, 0, 0, 0 }}, 
  { kCTLUP, { 27, 91, 49, 59, 53, 65, 0, 0 }}, 
  { kCTLDOWN, { 27, 79, 98, 0, 0, 0, 0, 0 }}, 
  { kCTLDOWN, { 27, 91, 49, 59, 53, 66, 0, 0 }}, 
  { kCTLLEFT, { 27, 79, 100, 0, 0, 0, 0, 0 }},
  { kCTLLEFT, { 27, 91, 49, 59, 53, 68, 0, 0 }},
  { kCTLRIGHT, { 27, 79, 99, 0, 0, 0, 0, 0 }},
  { kCTLRIGHT, { 27, 91, 49, 59, 53, 67, 0, 0 }},
  { kCTLX, { 24, 0, 0, 0, 0, 0, 0, 0 }},
  { kINS, { 27, 91, 50, 126, 0, 0, 0, 0 }}, 
  { kDEL, { 27, 91, 51, 126, 0, 0, 0, 0 }}, 
  { kPGUP, { 27, 91, 53, 126, 0, 0, 0, 0 }}, 
  { kPGDOWN, { 27, 91, 54, 126, 0, 0, 0, 0 }},
  { kHOME, { 27, 91, 55, 126, 0, 0, 0, 0 }},
  { kHOME, { 27, 91, 72, 0, 0, 0, 0, 0 }},
  { kEND, { 27, 91, 56, 126, 0, 0, 0, 0 }},
  { kEND, { 27, 91, 70, 0, 0, 0, 0, 0 }},
  { kALTC, { 27, 99, 0, 0, 0, 0, 0, 0 }},
  { kALTA, { 27, 97, 0, 0, 0, 0, 0, 0 }},
  { kALTS, { 27, 115, 0, 0, 0, 0, 0, 0 }},
};

bool_t next_key_code(keysym_t *s) {
  int i;

  if( SLang_input_pending(0) > 0 ) {
    memset(s, 0, sizeof(keysym_t));
    s->code[0] = SLang_getkey();

    if( s->code[0] < 32 ) { /* i > 0 */
      /* see if we recognize it */
      for(i = 1; i < KCLEN; i++) {
	s->code[i] = ( SLang_input_pending(0) > 0 ) ? SLang_getkey() : 0;
	if( s->code[i] == 0 ) { break; }
      }
      for(i = 0; i < sizeof(keysyms)/sizeof(keysym_t); i++) {
	if( memcmp(s->code, keysyms[i].code, sizeof(keycode_t)) == 0 ) {
	  s->sym = keysyms[i].sym;
	  return TRUE;
	}
      }
    } else if( (s->code[0] >= 32) && (s->code[0] < 128) ) {
      s->sym = kNORMAL;
      return TRUE;
    } else if( s->code[0] >= 128 ) {
      s->sym = kMULTIB;
      return TRUE;
    }
  }
  s->sym = kNOTHING;
  return FALSE;
}

void print_keysym(keysym_t *s) {
  int i;
  if( s->code[0] > 0 ) {
    SLsmg_set_color(3);
    SLsmg_gotorc(0,0);
    SLsmg_printf("keysym ");
    for(i = 0; i < KCLEN; i++) {
      SLsmg_printf("%d ", s->code[i]);
    }
    SLsmg_erase_eol();
    SLsmg_refresh();
  }
}

bool_t messageyn(char *msg) {
  keysym_t k;

  SLsmg_set_color(3);
  SLsmg_gotorc(disp.num_rows - 1, 0);
  SLsmg_printf("%s", msg);
  SLsmg_erase_eol();
  SLsmg_refresh();

  do {
    usleep(50); /* be polite */
    next_key_code(&k);
  } while(k.sym == kNOTHING);

  return ((k.sym == kNORMAL) &&
	  ((k.code[0] == 'y') || (k.code[0] == 'Y'))) ? TRUE : FALSE;
}

bool_t grow_single_box(box_t *box, imagebuffer_t *ib, int GROWF) {
  box_t box2;
  int delta;
  if( box && ib ) {
    box2.left = MAX(0,MIN(box->left, box->right) - GROWF);
    box2.right = MIN(ib->width,MAX(box->left, box->right) + GROWF);
    box2.bottom = MAX(0,MIN(box->top, box->bottom) - GROWF);
    box2.top = MIN(ib->height,MAX(box->top, box->bottom) + GROWF);

    if(simplecrop_bw_imagebuffer(ib,
				 box2.left,
				 ib->height - box2.top,
				 box2.right,
				 ib->height - box2.bottom, 0) ) {

      delta = 
	(abs(box->left - ib->crop.xmin) + 
	 abs(box->right - ib->crop.xmax) +
	 abs(box->top - ib->height + ib->crop.ymax) +
	 abs(box->bottom - ib->height + ib->crop.ymin));

      /* don't crop if the box contains nothing */
      if( ib->crop.coverage > 0 ) {
	box->left = ib->crop.xmin;
	box->right = ib->crop.xmax;
	box->top = ib->height - ib->crop.ymax; 
	box->bottom = ib->height - ib->crop.ymin;
      }

      return (delta > 0) && (ib->crop.coverage > 0) ? TRUE : FALSE;
    }
  }
  return FALSE;
}

bool_t grow_all_boxes(boxlist_t *bl, imagebuffer_t *ib, int GROWF) {
  int i;
  if( bl && ib ) {
    SLsmg_gotorc(disp.num_rows - 1, 0);
    SLsmg_printf("growing all boxes...");
    SLsmg_erase_eol();
    SLsmg_refresh();
    for(i = 0; i < bl->count; i++) {
      grow_single_box(&bl->items[i], ib, GROWF);
    }
  }
  return TRUE; 
}

bool_t shrink_box(box_t *box, imagebuffer_t *ib, int s) {
  box_t box2;

  if( box && ib ) {
    
    box2.left = (box->left + box->right)/2 - s;
    box2.right = (box->left + box->right)/2 + s;
    /* note bottom/top reversal */
    box2.bottom = (box->top + box->bottom)/2 + s;
    box2.top = (box->top + box->bottom)/2 - s;

    box->left = MAX(0,box2.left);
    box->right = MIN(ib->width,box2.right);
    box->bottom = MIN(ib->height,box2.bottom);
    box->top = MAX(0,box2.top);
    return TRUE;
  }
  return FALSE;
}

/* this grows more accurately, but is slower */
bool_t grow_single_box2(box_t *box, imagebuffer_t *ib, int GROWF) {
  int i;
  if( box && ib ) {

    for(i = 0; i < GROWF; i++) {

      if( !grow_single_box(box, ib, 2) ) {
	break;
      }

    }
    return TRUE;
  }
  return FALSE;
}

bool_t grow_all_boxes2(boxlist_t *bl, imagebuffer_t *ib, int GROWF) {
  int i;
  if( bl && ib ) {
    SLsmg_gotorc(disp.num_rows - 1, 0);
    SLsmg_printf("growing all boxes...");
    SLsmg_erase_eol();
    SLsmg_refresh();
    for(i = 0; i < bl->count; i++) {
      grow_single_box2(&bl->items[i], ib, GROWF);
    }
  }
  return TRUE; 
}

bool_t run_boxeditor(boxeditor_t *e) {
  bool_t done = FALSE;
  bool_t flush = FALSE;

  int xp, yp;
  xtwin_event_t ev;
  box_t *box, box2;
  keysym_t k = { kNOTHING };
  int keypress = 0;

  /* readline stuff */
  char *buf;
  SLrline_Type *rli;
  unsigned int len = 0;

  /* uncomment this for debug */
  /* u_options |= FL_KEYSIMS;  */

  while( !done ) {

    /* this is eating too much cpu! someday I'll have
       to investigate :) */

    ev = next_event_xtwin(&e->xtw);
    next_key_code(&k);

    if( (ev == NO_EVENTS) && (k.sym == kNOTHING) ) {
      if( e->dirty_graph && flush ) {
	/* this might take a while */
	SLsmg_set_color(2);

	SLsmg_gotorc(disp.num_rows - 1, 0);
	SLsmg_printf("rendering...");
	SLsmg_erase_eol();
	SLsmg_refresh();

	render_boxlist(e->bl, e->ib, 1);
	e->dirty_graph = FALSE;
	flush = FALSE;
	
      } else {
	usleep(200); /* be polite */
      }
    } 

    box = ((e->bl->count > 0) ?
	   &(e->bl->items[disp.highlighted + disp.first_visible]) :
	   NULL);

    if( e->dirty_list ) {
      redraw_boxeditor(e); 
    }

    if( (ev == READY_EVENT) || e->dirty_list ) {
      redraw_xtwin(&e->xtw, e->ib, box);
    }

    e->dirty_list = FALSE;

    if( u_options & FL_KEYSIMS ) {
      print_keysym(&k);
    }

    if( k.sym == kNOTHING ) {
      continue; /* nothing happened */
    }


    /* if user presses a single key, just update symbol,
       if user presses a series of keys or an input method
       announcer, bring up an editor */

    if( (k.sym == kNORMAL) || (k.sym == kMULTIB) ) {
      if( !keypress && (k.sym == kNORMAL) ) {
	memset(box->symbol, 0, sizeof(box->symbol));	
	keypress = box->symbol[0] = k.code[0];
      } else {
	SLang_ungetkey(k.code[0]);
	if( (k.sym == kNORMAL) && keypress ) {
	  SLang_ungetkey(keypress);
	}

	rli = SLrline_open(SYMLEN, SL_RLINE_BLINK_MATCH|SL_RLINE_UTF8_MODE);
	if( rli ) {
	  len = SYMLEN + 1;

/* 	box2.symbol[0] = keypress; */
/* 	box2.symbol[1] = k.code[0]; */
/* 	box2.symbol[2] = '\0'; */
/* 	SLrline_set_line(rli, (char *)box2.symbol); */
	  SLrline_set_line(rli, "");

	  SLtt_set_cursor_visibility(1);
	  SLsmg_normal_video();

	  /* status line */
	  SLsmg_set_color(2);

	  SLsmg_gotorc(disp.num_rows - 1, 0);
	  SLsmg_printf("ENTER: finish editing");
	  SLsmg_erase_eol();

	  /* important: gotorc is needed so
	     that SLrline_read_line happens
	     where we want it */

	  SLsmg_gotorc(disp.highlighted + 1, 0);
	  SLsmg_printf("        \t");
	  SLsmg_gotorc(disp.highlighted + 1, 2);
	  SLsmg_refresh();
	  buf = SLrline_read_line(rli, " ", &len);
	  if( buf ) {
	    memset(box->symbol, 0, sizeof(box->symbol));
	    strncpy((char *)box->symbol, buf, SYMLEN);
	    /* add_history(strdup(buf)); */
	  }
	  SLfree(buf);
	  SLtt_set_cursor_visibility(0);
	}
	SLrline_close(rli);
	keypress = 0;
      }
      e->dirty_list = TRUE;
    } else { /* a special key */
      keypress = 0;
    }

    /* now handle any special keys we recognize */

    switch(k.sym) {
    case kESC:
      if( messageyn("ESC was pressed. Discard all changes (y/n)?") ) {
	return FALSE;
      }
      e->dirty_list = TRUE;
      break;
    case kUP:
    case kLEFT:
      if( disp.highlighted > 0 ) {
	disp.highlighted--;
      } else {
	if( disp.first_visible > 1 ) 
	  { disp.first_visible -= 1; }
	else
	  { disp.first_visible = 0; }
      }
      e->dirty_list = TRUE;
      flush = TRUE;
      break;
    case kDOWN:
    case kRIGHT:
      if( e->bl->count > 0 ) {
	if( disp.highlighted < (e->bl->count - disp.first_visible - 1) ) {
	  if( disp.highlighted < (disp.num_rows - 4) ) {
	    disp.highlighted++;
	  } else {
	    if( (disp.first_visible += 1) >= e->bl->count ) 
	      { disp.first_visible = e->bl->count - 1; }
	  }
	}
      } else {
	disp.highlighted = disp.first_visible = 0;
      }
      e->dirty_list = TRUE;
      flush = TRUE;
      break;
    case kALTUP:
      if( box && 
	  (box->top < e->ib->height) && (box->bottom < e->ib->height) ) {
	box->top++;
	box->bottom++;
	e->dirty_list = TRUE;
	e->dirty_graph = TRUE;
      }
      break;
    case kALTDOWN:
      if( box && (box->bottom > 0) && (box->top > 0) ) {
	box->top--;
	box->bottom--;
	e->dirty_list = TRUE;
	e->dirty_graph = TRUE;
      }
      break;
    case kALTLEFT:
      if( box && (box->left > 0) && (box->right > 0) ) {
	box->left--;
	box->right--;
	e->dirty_list = TRUE;
	e->dirty_graph = TRUE;
      }
      break;
    case kALTRIGHT:
      if( box && 
	  (box->right < e->ib->width) && (box->left < e->ib->width) ) {
	box->left++;
	box->right++;
	e->dirty_list = TRUE;
	e->dirty_graph = TRUE;
      }
      break;
    case kCTLUP:
      if( box && 
	  (box->top > 0) && (box->bottom < e->ib->height) &&
	  (abs(box->top - 1 - box->bottom - 1) > 1) ) {
	box->top--;
	box->bottom++;
	e->dirty_list = TRUE;
	e->dirty_graph = TRUE;
      }
      break;
    case kCTLDOWN:
      if( box && 
	  (box->top < e->ib->height) && (box->bottom > 0) &&
	  (abs(box->top + 1 - box->bottom + 1) > 1) ) {
	box->top++;
	box->bottom--;
	e->dirty_list = TRUE;
	e->dirty_graph = TRUE;
      }
      break;
    case kCTLLEFT:
      if( box && 
	  (box->left < e->ib->width) && (box->right > 0) &&
	  (abs(box->left + 1 - box->right + 1) > 1) ) {
	box->left++;
	box->right--;
	e->dirty_list = TRUE;
	e->dirty_graph = TRUE;
      }
      break;
    case kCTLRIGHT:
      if( box && 
	  (box->left > 0) && (box->right < e->ib->width) &&
	  (abs(box->left - 1 - box->right - 1) > 1) ) {
	box->left--;
	box->right++;
	e->dirty_list = TRUE;
	e->dirty_graph = TRUE;
      }
      break;
    case kCTLX:
      done = TRUE;
      break;
    case kINS:
      xp = rand() % 20 - 10;
      yp = rand() % 20 - 10;
      memset(box2.symbol, 0, sizeof(box2.symbol));
      box2.left = box->left + xp;
      box2.right = box->right + xp;
      box2.top = box->top + yp;
      box2.bottom = box->bottom + yp;
      insert_boxlist(e->bl, &box2, disp.highlighted + disp.first_visible);
      e->dirty_list = TRUE;
      e->dirty_graph = TRUE;
      break;
    case kDEL:
      remove_boxlist(e->bl, disp.highlighted + disp.first_visible);
      e->dirty_list = TRUE;
      e->dirty_graph = TRUE;
      break;
    case kPGUP:
      if( disp.first_visible > disp.num_rows )
	{ disp.first_visible -= disp.num_rows; }
      else
	{ disp.first_visible = 0; disp.highlighted = 0; }

      /* assert emails.num_limited >= disp.first_visible */
      if( disp.highlighted > (e->bl->count - disp.first_visible) ) {
	disp.highlighted = (e->bl->count - disp.first_visible);
      }
      e->dirty_list = TRUE;
      break;
    case kPGDOWN:
      if( e->bl->count > 0 ) {
	if( (disp.first_visible += disp.num_rows) >= e->bl->count ) 
	  { disp.first_visible = e->bl->count - 1; }

	if( disp.highlighted > (e->bl->count - disp.first_visible) ) {
	  disp.highlighted = (e->bl->count - disp.first_visible) - 1;
	}
      } else {
	disp.highlighted = disp.first_visible = 0;
      }
      e->dirty_list = TRUE;
      break;
    case kHOME:
      disp.first_visible = 0;
      e->dirty_list = TRUE;
      break;
    case kEND:
      disp.first_visible = (e->bl->count > 0) ? e->bl->count - 1 : 0;
      e->dirty_list = TRUE;
      break;
    case kALTC:
      /* grow the box a few units */
      if( grow_single_box2(box, e->ib, 5) ) {
	e->dirty_list = TRUE;
	e->dirty_graph = TRUE;
      }
      break;
    case kALTS:
      /* shrink the box towards the center */
      if( shrink_box(box, e->ib, 5) ) {
	e->dirty_list = TRUE;
	e->dirty_graph = TRUE;
      }
      break;
    case kALTA:
      /* grow ALL boxes */
      if( messageyn("Alt-A Pressed. Crop all boxes at once (y/n)?") ) {
	if( grow_all_boxes2(e->bl, e->ib, 5) ) {
	  e->dirty_graph = TRUE;
	  flush = TRUE;
	}
	e->dirty_list = TRUE;
      }
      break;
    default:
      /* nothing */
      break;
    }

  }

  return TRUE;  
}

bool_t gotoxy_boxeditor(boxeditor_t *e, long int x, long int y) {
  int i;
  box_t *b;
  unsigned int xx, yy;

  if( e && (x > -1) && (y > -1) ) {
    xx = (unsigned int)x;
    yy = (unsigned int)y;

    for(i = 0; i < e->bl->count; i++) {
      b = &(e->bl->items[i]);
      if( ((b->left == xx) && (b->top == yy)) ||
	  ((b->right == xx) && (b->bottom == yy)) ) {
	disp.first_visible = i;
	disp.highlighted = 0;
	return TRUE;
      }
    }

  }
  return FALSE;
}

