/* 
 * Copyright (C) 2008 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifndef IMAGEBUFFER_H
#define IMAGEBUFFER_H

#include "common.h"
#include "ppmwriter.h"


/* The bits are numbered starting from zero. To keep things consistent 
 * the BW macros read and set individual bits, which represent
 * black pixels. the GS macros read and set 4 bit wide sections
 * which represent grayscale pixels. No error checking is done,
 * so make sure bitmap is big enough etc. If bitmap is 16 bits,
 * then we can address 16 pixels, or 4 grayscale pixels
 */

#define MAX_GS 16

#define READBW(bitmap,b) (( (bitmap) & (1<<(b)) ) > 0)
#define READGS(bitmap,b) ( ( (bitmap) & (0xF<<(4*(b))) )>>(4*(b))    )

#define WRITEBW(bitmap,b,v) bitmap = (( (bitmap) & ~(1<<(b))) + ((v)<<(b)))
#define WRITEGS(bitmap,b,v) bitmap = (( (bitmap) & ~(0xF<<(4*(b)))) + ((v)<<(4*(b))))
#define SETBW(bitmap,b) bitmap |= (1<<(b))
#define SETGS(bitmap,b) bitmap |= (0xF<<(4*(b)))

#define CLEARBW(bitmap,b) bitmap &= ~(1<<(b))
#define CLEARGS(bitmap,b) bitmap &= ~(0xF<<(4*(b)))

/* the imagebuffer contains the pixel data from all the scanned pages
 * as a frequency histogram. It also contains a few image planes so we
 * can copy an image into it, and perform some transformations. The
 * structure is packed into as few bytes as practicable to speed
 * things up.
 *
 * The standard planes are as follows:
 * bw plane 0 is for the unmodified bw image
 * bw plane 1 is a scratchpad 
 * bw plane 2 is a scratchpad 
 * gs plane 0 is for an unmodified gs image (16 colours max)
 * gs plane 1 is a scratchpad
 * gs plane 2 is a scratchpad
 * gs plane 3 is a scracthpad
 */ 
typedef struct {
  uint16_t pixl;
  uint16_t pixf;
} PACK_STRUCTS imagepoint_t;

#define PIXFREQ 16
/* this is the capacity of pixf  */
#define MAXIMAGES (1<<(PIXFREQ))

/* this is for histogram equalizer */
#define MAX_HISTEQ 255

typedef struct {
  size_t width;
  size_t height;
  struct {
    int x;
    int y;
  } pos;
  imagepoint_t *image;
  size_t image_size;
  struct {
    unsigned int min;
    unsigned int max;
    unsigned long cdf[MAX_GS];
    unsigned int histeq[MAX_GS];
  } colours;
  struct {
    int xmin;
    int ymin;
    int xmax;
    int ymax;
    long coverage;
  } crop;
} imagebuffer_t;

bool_t create_imagebuffer(imagebuffer_t *ib);
bool_t clear_imagebuffer(imagebuffer_t *ib);
bool_t free_imagebuffer(imagebuffer_t *ib);
bool_t grow_imagebuffer(imagebuffer_t *ib, int w, int h);

bool_t seekpos_imagebuffer(imagebuffer_t *ib, int x, int y);


bool_t box_bw_imagebuffer(imagebuffer_t *ib, 
			  int xmin, int ymin, int xmax, int ymax,
			  int plane, int val);
bool_t fill_bw_imagebuffer(imagebuffer_t *ib, 
			   int xmin, int ymin, int xmax, int ymax,
			   int plane, int val);
bool_t mv_bw_imagebuffer(imagebuffer_t *ib,
			 int xmin, int ymin, int xmax, int ymax,
			 int from, int to);

bool_t simplecrop_bw_imagebuffer(imagebuffer_t *ib,
				 int xmin, int ymin, int xmax, int ymax,
				 int plane);
bool_t crop_bw_imagebuffer(imagebuffer_t *ib, int level, int dp, int sp);


bool_t save_color_imagebuffer(imagebuffer_t *ib, ppmwriter_t *p,
			      imagetype_t typ, int plane);

bool_t partfill_bw_imagebuffer(imagebuffer_t *ib, byte_t *buffer, 
			       int numbits, int width, int height,
			       int plane);


bool_t fillhist_imagebuffer(imagebuffer_t *ib, int plane);

void render_pixel_imagebuffer(imagepoint_t *point, int plane,
			      byte_t *red, byte_t *green, byte_t *blue);

#endif
