/* 
 * Copyright (C) 2008 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "mask.h"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

bool_t create_mask(mask_t *m) {
  if( m ) {
    return TRUE;
  }
  return FALSE;
}

bool_t free_mask(mask_t *m) {
  if( m ) {
    return TRUE;
  }
  return FALSE;
}


void printimage_pixl01(imagebuffer_t *ib, mask_t *m) {
  int x, y;
  printf("---\n");
  for(y = 0; y < ib->height; y++) {
    for(x = 0; x < ib->width; x++) {
      printf(READBW(ib->image[y*ib->width + x].pixl,0) ? "*" : " ");
    }
    printf("|");
    for(x = 0; x < ib->width; x++) {
      printf(READBW(ib->image[y*ib->width + x].pixl,1) ? "*" : " ");
    }
    printf("|");
    for(x = 0; x < ib->width; x++) {
      printf(((x <= m->r.xmax) && (x >= m->r.xmin) && (y <= m->r.ymax) && (y >= m->r.ymin)) ? "#" : " ");
    }
    printf("|\n");
  }
  printf("---\n");
}


/* start at (x0,y0) and mask the connected component
 which contains it. Input taken from pixl0, output
 goes to pixl1. 

 Here's what we do: starting from the single point (x0,y0), we make
 bigger and bigger rectangles. Each rectangle goes from (xmin,ymin)
 to (xmax,ymax) but is clipped by the full image.

 Each pixel has 8 possible neighbours (ie diagonals are possible).
 For each pixel on the rectangle border, the colour is changed to 
black if:
 1) it is black in the pixl0 plane and
 2) it is already black or
    one of its neighbours inside the rectangle is black

 When a rectangle border is finished, we do the next bigger rectangle.
 We never do a border which touches the full picture border (ie when
 xmin = 0, or xmax = ib->width, or ymin = 0 or ymax = ib->height)

 To see if the algorithm should stop, we count the total number of 
 pixel changes for each rectangle, this is easy since a pixel can
 only go from 0 to 1 (or 0 to 0) so we make a telescoping sum
 of differences by the before and after values of each pixel.

*/

#define PIXEL0(p) (READBW((p)->pixl,from))
#define PIXEL1(p) (READBW((p)->pixl,to))
#define SETPIXEL1(p,v) WRITEBW((p)->pixl,to,v)

bool_t make_component_bw_mask(mask_t *m, imagebuffer_t *ib, int x0, int y0,
			      int from, int to) {
  register int i, b;
  imagepoint_t *pointA, *pointB;

  if( m && ib && ib->image ) {

    SETPIXEL1(ib->image + y0 * ib->width + x0,to);
    m->r.xmin = m->r.xmax = x0;
    m->r.ymin = m->r.ymax = y0;

    m->c.xmin = m->c.xmax = x0;
    m->c.ymin = m->c.ymax = y0;

    b = 1;  /* counts the number of changed pixels */
    while( (b > 0) && 
	   ((m->r.xmax - m->r.xmin < ib->width) || 
	    (m->r.ymax - m->r.ymin < ib->height)) ) {
      m->r.xmin = MAX(0, m->r.xmin - 1);
      m->r.xmax = MIN(ib->width - 1, m->r.xmax + 1);
      m->r.ymin = MAX(0, m->r.ymin - 1);
      m->r.ymax = MIN(ib->height - 1, m->r.ymax + 1);
      /* count the number of modified pixels */
      b = 0; 

      if( m->r.ymin > 0 ) {
	/* top row */
	pointA = ib->image + m->r.ymin * ib->width + m->r.xmin;
	pointB = pointA + ib->width;

	/* leftmost pixel */
	if( PIXEL1(pointA) == 0 ) {
	  b -= PIXEL1(pointA);
	  SETPIXEL1(pointA,
		    (PIXEL0(pointA) & 
		     ( PIXEL1(pointA) | PIXEL1(pointA + 1) |
		       PIXEL1(pointB) | PIXEL1(pointB + 1))));
	  b += PIXEL1(pointA);
	}
	pointA++;
	pointB++;

	/* other pixels */
	for(i = m->r.xmin + 1; i < m->r.xmax; i++) {
	  if( PIXEL1(pointA) == 0 ) {
	    b -= PIXEL1(pointA);
	    SETPIXEL1(pointA,
		     (PIXEL0(pointA) & 
		      (PIXEL1(pointA - 1) | PIXEL1(pointA) | PIXEL1(pointA + 1) |
		       PIXEL1(pointB - 1) | PIXEL1(pointB) | PIXEL1(pointB + 1))));
	    b += PIXEL1(pointA);
	  }
	  pointA++;
	  pointB++;
	}

	/* rightmost pixel */
	if( PIXEL1(pointA) == 0 ) {
	  b -= PIXEL1(pointA);
	  SETPIXEL1(pointA,
		    (PIXEL0(pointA) & 
		     ( PIXEL1(pointA - 1) | PIXEL1(pointA) |
		       PIXEL1(pointB - 1) | PIXEL1(pointB) )));
	  b += PIXEL1(pointA);
	}
	pointA++;
	pointB++;

	/* update crop bounds */
	if( b > 0 ) {
	  m->c.ymin = m->r.ymin;
	}
      }

      if( m->r.xmin > 0 ) {
	/* left side */
	for(i = m->r.ymin + 1; i < m->r.ymax; i++) {
	  pointA = ib->image + i * ib->width + m->r.xmin;
	  if( PIXEL1(pointA) == 0 ) {
	    b -= PIXEL1(pointA);
	    SETPIXEL1(pointA,
	      (PIXEL0(pointA) & 
	       (PIXEL1(pointA - ib->width) | PIXEL1(pointA - ib->width + 1) |
		PIXEL1(pointA) | PIXEL1(pointA + 1) |
		PIXEL1(pointA + ib->width) | PIXEL1(pointA + ib->width + 1))));
	    b += PIXEL1(pointA);
	  }
	}
	/* update crop bounds */
	if( b > 0 ) {
	  m->c.xmin = m->r.xmin;
	}
      }
      if( m->r.xmax < ib->width ) {
	/* right side */
	for(i = m->r.ymin + 1; i < m->r.ymax; i++) {
	  pointA = ib->image + i * ib->width + m->r.xmax;
	  if( PIXEL1(pointA) == 0 ) {
	    b -= PIXEL1(pointA);
	    SETPIXEL1(pointA,
	      (PIXEL0(pointA) & 
	       (PIXEL1(pointA - ib->width) | PIXEL1(pointA - ib->width + 1) |
		PIXEL1(pointA) | PIXEL1(pointA - 1) |
		PIXEL1(pointA + ib->width) | PIXEL1(pointA + ib->width + 1))));
	    b += PIXEL1(pointA);
	  }
	}
	/* update crop bounds */
	if( b > 0 ) {
	  m->c.xmax = m->r.xmax;
	}
      }

      if( m->r.ymax < ib->height ) {
	/* bottom row */
	pointA = ib->image + m->r.ymax * ib->width + m->r.xmin;
	pointB = pointA - ib->width;

	/* leftmost pixel */
	if( PIXEL1(pointA) == 0 ) {
	  b -= PIXEL1(pointA);
	  SETPIXEL1(pointA,
		    (PIXEL0(pointA) & 
		     ( PIXEL1(pointA) | PIXEL1(pointA + 1) |
		       PIXEL1(pointB) | PIXEL1(pointB + 1))));
	  b += PIXEL1(pointA);
	}
	pointA++;
	pointB++;

	/* other pixels */
	for(i = m->r.xmin + 1; i < m->r.xmax; i++) {
	  if( PIXEL1(pointA) == 0 ) {
	    b -= PIXEL1(pointA);
	    SETPIXEL1(pointA,
	      (PIXEL0(pointA) & 
	       (PIXEL1(pointA - 1) | PIXEL1(pointA) | PIXEL1(pointA + 1) |
		PIXEL1(pointB - 1) | PIXEL1(pointB) | PIXEL1(pointB + 1))));
	    b += PIXEL1(pointA);
	  }
	  pointA++;
	  pointB++;
	}

	/* rightmost pixel */
	if( PIXEL1(pointA) == 0 ) {
	  b -= PIXEL1(pointA);
	  SETPIXEL1(pointA,
		    (PIXEL0(pointA) & 
		     (PIXEL1(pointA - 1) | PIXEL1(pointA) |
		      PIXEL1(pointB - 1) | PIXEL1(pointB) )));
	  b += PIXEL1(pointA);
	}
	pointA++;
	pointB++;

	/* update crop bounds */
	if( b > 0 ) {
	  m->c.ymax = m->r.ymax;
	}
      }

      /* printimage(ib,m); */

    }
    return TRUE;
  }
  return FALSE;
}

